/* hw/s3c2410x_irq.c
 *
 * Samsung S3C2410X emulation
 *
 * Copyright 2006 Daniel Silverstone and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2
 */

#include "vl.h"
#include "s3c2410x.h"

static uint32_t s3c2410x_irq_main_level, s3c2410x_irq_subsrc_level;
static uint32_t s3c2410x_irq_reg[8];
static CPUState *cpu_env;
#define IR(X) s3c2410x_irq_reg[X]

static void
__s3c2410x_percolate_interrupt(void)
{
  /* Take the status of the srcpnd register, percolate it through, raise to CPU if necessary */
  uint32_t ints = (s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] & 
                   ~s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTMSK]);
  int fsb = ffs(ints);
  
  /* TODO: Priority encoder could go here */
  if (ints & s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTMOD]) {
    /* Detected a FIQ */
    S3C2410X_DBF("IRQ: Setting FIQ on CPU\n");
    cpu_interrupt(cpu_env, CPU_INTERRUPT_FIQ);
    return;
  } else {
    /* No FIQ here today */
    cpu_reset_interrupt(cpu_env, CPU_INTERRUPT_FIQ);
  }
  /* No FIQ, do we have an IRQ */
  if (fsb) {
    if ((s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTPND] == 0) ||
        (s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTPND] > 1<<(fsb-1))) {
      /* Current INTPND is lower priority than fsb of ints (or empty) */
      s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTPND] = 1<<(fsb-1);
      s3c2410x_irq_reg[CPU_S3C2410X_IRQ_OFFSET] = fsb-1;
    }
  } else {
    /* No FSB, thus no IRQ, thus nothing to see yet */
  }
  if (s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTPND] != 0) {
    S3C2410X_DBF("IRQ: Setting IRQ on CPU\n");
    cpu_interrupt(cpu_env, CPU_INTERRUPT_HARD);
  } else {
    cpu_reset_interrupt(cpu_env, CPU_INTERRUPT_HARD);
  }
}

static void
s3c2410x_percolate_subsrc_interrupt(void)
{
  uint32_t ints;
  
  s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= s3c2410x_irq_main_level;
  s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND] |= s3c2410x_irq_subsrc_level;
  
  ints = (s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND] &
          ~s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTSUBMSK]);
   
  /* If UART0 has asserted, raise that */
  S3C2410X_DBF("SUBSRCPND 0x%08x\n", s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND]);
  S3C2410X_DBF("INTSUBMSK 0x%08x\n", s3c2410x_irq_reg[CPU_S3C2410X_IRQ_INTSUBMSK]);
  S3C2410X_DBF(" => Pending Subsource interrupts: %08x\n", ints);
  if( ints & 0x7 ) {
    S3C2410X_DBF("IRQ: Percolating UART0 subsource\n");
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= (1<<28);
  }
  /* Ditto UART1 */
  if( ints & 0x7<<3 ) s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= (1<<23);
  /* Ditto UART2 */
  if( ints & 0x7<<6 ) s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= (1<<15);
  /* And percolate it through */
  __s3c2410x_percolate_interrupt();
}

static void
s3c2410x_irq_write_f(void *opaque_ignored, target_phys_addr_t addr_, uint32_t value)
{
  int addr = (addr_ >> 2) & 7;
  S3C2410X_DBF("Write IRQ[%02x] = %08x\n",addr<<2, value);
  if (addr == CPU_S3C2410X_IRQ_SRCPND ||
      addr == CPU_S3C2410X_IRQ_INTPND ||
      addr == CPU_S3C2410X_IRQ_SUBSRCPND) {
    s3c2410x_irq_reg[addr] &= ~value;
  } else
    s3c2410x_irq_reg[addr] = value;
  /* Start at the subsrc irqs and percolate from there */
  s3c2410x_percolate_subsrc_interrupt();
}

static uint32_t
s3c2410x_irq_read_f(void *opaque_ignored, target_phys_addr_t addr_)
{
  int addr = (addr_ >> 2) & 0x7;
  S3C2410X_DBF("Read IRQ[%02x] => 0x%08x\n",addr<<2, s3c2410x_irq_reg[addr]);
  return s3c2410x_irq_reg[addr];
}


static CPUReadMemoryFunc *s3c2410x_irq_read[] = {
  &s3c2410x_irq_read_f,
  &s3c2410x_irq_read_f,
  &s3c2410x_irq_read_f,
};

static CPUWriteMemoryFunc *s3c2410x_irq_write[] = {
  &s3c2410x_irq_write_f,
  &s3c2410x_irq_write_f,
  &s3c2410x_irq_write_f,
};

void
s3c2410x_irq_init(CPUState *env)
{
  /* Samsung S3C2410X IRQ magic badger.
   *
   * Interrupt controller scary ARGH
   */
  int tag = cpu_register_io_memory(0, s3c2410x_irq_read, s3c2410x_irq_write, env);
  cpu_register_physical_memory(CPU_S3C2410X_IRQ_BASE, 8*4, tag);
  cpu_env = env; /* UGLY, Only one IRQ controller, only one CPU Damnit */
  IR(CPU_S3C2410X_IRQ_SRCPND) = 0x00;
  IR(CPU_S3C2410X_IRQ_INTMOD) = 0x00;
  IR(CPU_S3C2410X_IRQ_INTMSK) = 0xFFFFFFFF;
  IR(CPU_S3C2410X_IRQ_PRIORITY) = 0x7F; /* Not that we use it */
  IR(CPU_S3C2410X_IRQ_INTPND) = 0x00;
  IR(CPU_S3C2410X_IRQ_OFFSET) = 0x00;
  IR(CPU_S3C2410X_IRQ_SUBSRCPND) = 0x00;
  IR(CPU_S3C2410X_IRQ_INTSUBMSK) = 0x7FF;
}

void
s3c2410x_irq_raise_interrupt_sif(void *opaque_ignored, int irq_num, int level)
{
  /* Raise irq_num */
  S3C2410X_DBF("IRQ: Set pulse for %d to %d\n", irq_num, level);
  if( level )
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= 1<<irq_num;
  s3c2410x_percolate_subsrc_interrupt();
}

void
s3c2410x_irq_raise_subsrc_interrupt_sif(void *opaque_ignored, int irq_num, int level)
{
  S3C2410X_DBF("IRQ: Subsrc pulse set %d to %d\n", irq_num, level);
  if( level ) {
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND] |= 1<<irq_num;
  }
  s3c2410x_percolate_subsrc_interrupt();
}

void
s3c2410x_irq_set_interrupt_level_sif(void *opaque_ignored, int irq_num, int level)
{
  S3C2410X_DBF("IRQ: Set level for %d to %d\n", irq_num, level);
  if( level ) {
    s3c2410x_irq_main_level |= 1<<irq_num;
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] |= 1<<irq_num;
  } else {
    s3c2410x_irq_main_level &= ~(1<<irq_num);
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SRCPND] &= ~(1<<irq_num);
  }
  s3c2410x_percolate_subsrc_interrupt();
}

void
s3c2410x_irq_set_subsrc_interrupt_level_sif(void *opaque_ignored, int irq_num, int level)
{
  S3C2410X_DBF("IRQ: Set subsrc level for %d to %d\n", irq_num, level);
  if( level ) {
    s3c2410x_irq_subsrc_level |= 1<<irq_num;
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND] |= 1<<irq_num;
  } else {
    s3c2410x_irq_subsrc_level &= ~(1<<irq_num);
    s3c2410x_irq_reg[CPU_S3C2410X_IRQ_SUBSRCPND] &= ~(1<<irq_num);
  }
  s3c2410x_percolate_subsrc_interrupt();
}
