/* hw/s3c2410x_serial.c
 *
 * Samsung S3C2410X Serial block
 *
 * Copyright 2006 Daniel Silverstone and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2
 */

#include "vl.h"
#include "s3c2410x.h"

typedef struct {
  uint32_t ulcon, ucon, ufcon, umcon, ubrdiv;
  unsigned char rx_byte;
  /* Byte is available to be read */
  int rx_available : 1;
  CharDriverState *chr;
  int port;
} s3c2410x_serial_dev;

static void 
s3c2410x_serial_write_f(void *opaque, target_phys_addr_t addr, uint32_t value)
{
  s3c2410x_serial_dev *s = opaque;
  int reg = addr & 0x3f;
  if (reg != CPU_S3C2410X_SERIAL_UTXH)
    S3C2410X_DBF("Serial[%d] Write of %08x to %08x (reg %d)\n", s->port, value, addr, reg);
  switch(reg) {
  case CPU_S3C2410X_SERIAL_ULCON:
    s->ulcon = value;
    break;
  case CPU_S3C2410X_SERIAL_UCON:
    s->ucon = value;
    if( s->ucon & 1<<9 ) {
      S3C2410X_DBF("Serial[%d] setting TX level interrupt due to ucon write\n", s->port);
      s3c2410x_irq_set_subsrc_interrupt_level(1+(s->port*3), 1);
    } else {
      S3C2410X_DBF("Serial[%d] clearing TX level interrupt due to ucon write\n", s->port);
      s3c2410x_irq_set_subsrc_interrupt_level(1+(s->port*3), 0);
    }
    if( !(s->ucon & 1<<8) ) {
      S3C2410X_DBF("Serial[%d] clearing RX level interrupt (if any) due to ucon write\n", s->port);
      s3c2410x_irq_set_subsrc_interrupt_level((s->port*3), 0);
    }
    break;
  case CPU_S3C2410X_SERIAL_UFCON:
    s->ufcon = value;
    break;
  case CPU_S3C2410X_SERIAL_UMCON:
    s->umcon = value;
    break;
  case CPU_S3C2410X_SERIAL_UTRSTAT:
    break;
  case CPU_S3C2410X_SERIAL_UERSTAT:
    break;
  case CPU_S3C2410X_SERIAL_UFSTAT:
    break;
  case CPU_S3C2410X_SERIAL_UMSTAT:
    break;
  case CPU_S3C2410X_SERIAL_UTXH: {
    unsigned char ch = value & 0xff;
    if (s->chr && ((s->ucon & 1<<5)==0))
      qemu_chr_write(s->chr, &ch, 1);
    else {
      s->rx_byte = ch;
      s->rx_available = 1;
      if( s->ucon & 1<<8 ) {
        S3C2410X_DBF("Serial[%d] Setting RX level interrupt due to loopback write\n", s->port);
        s3c2410x_irq_set_subsrc_interrupt_level(s->port*3, 1);
      } else
        s3c2410x_irq_raise_subsrc_interrupt(s->port*3);
    }
    if( s->ucon & 1<<9 ) {
      S3C2410X_DBF("Serial[%d] Setting TX level interrupt due to byte write\n", s->port);
      s3c2410x_irq_set_subsrc_interrupt_level(1+(s->port*3), 1);
    } else
      s3c2410x_irq_raise_subsrc_interrupt(1+(s->port*3));
    break;
  }
  case CPU_S3C2410X_SERIAL_URXH:
    break;
  case CPU_S3C2410X_SERIAL_UBRDIV:
    s->ubrdiv = value;
    break;
  default:
    break;
  };
}

static uint32_t
s3c2410x_serial_read_f(void *opaque, target_phys_addr_t addr)
{
  s3c2410x_serial_dev *s = opaque;
  int reg = addr & 0x3f;
  if (reg != CPU_S3C2410X_SERIAL_URXH && 
      reg != CPU_S3C2410X_SERIAL_UFCON &&
      reg != CPU_S3C2410X_SERIAL_UFSTAT)
    S3C2410X_DBF("Serial[%d] Read of %08x (reg %d)\n", s->port, addr, reg);
  switch(reg) {
  case CPU_S3C2410X_SERIAL_ULCON:
    return s->ulcon;
  case CPU_S3C2410X_SERIAL_UCON:
    return s->ucon;
  case CPU_S3C2410X_SERIAL_UFCON:
    return s->ufcon & ~0x8; /* bit 3 is reserved, must be zero */
  case CPU_S3C2410X_SERIAL_UMCON:
    return s->umcon & 0x11; /* Rest are reserved, must be zero */
  case CPU_S3C2410X_SERIAL_UTRSTAT:
    return 6 | s->rx_available; /* TX always clear, RX when available */
  case CPU_S3C2410X_SERIAL_UERSTAT:
    return 0; /* Later, break detect comes in here */
  case CPU_S3C2410X_SERIAL_UFSTAT:
    return s->rx_available; /* TXFIFO, always empty, RXFIFO 0 or 1 bytes */
  case CPU_S3C2410X_SERIAL_UMSTAT:
    return 0;
  case CPU_S3C2410X_SERIAL_UTXH:
    return 0;
  case CPU_S3C2410X_SERIAL_URXH:
    s->rx_available = 0;
    if( s->ucon & 1<<8 ) {
      S3C2410X_DBF("Serial[%d] clearing RX level interrupt due to byte read\n", s->port);
      s3c2410x_irq_set_subsrc_interrupt_level(s->port*3, 0);
    }
    return s->rx_byte;
  case CPU_S3C2410X_SERIAL_UBRDIV:
    return s->ubrdiv;
  default:
    return 0;
  };
}

static CPUReadMemoryFunc *s3c2410x_serial_read[] = {
  &s3c2410x_serial_read_f,
  &s3c2410x_serial_read_f,
  &s3c2410x_serial_read_f,
};

static CPUWriteMemoryFunc *s3c2410x_serial_write[] = {
  &s3c2410x_serial_write_f,
  &s3c2410x_serial_write_f,
  &s3c2410x_serial_write_f,
};


static void s3c2410x_serial_event(void *opaque, int event)
{
  /*
  s3c2410x_serial_dev *s = opaque;
    if (event == CHR_EVENT_BREAK)
    serial_receive_break(s);
  */
}

static int 
s3c2410x_serial_can_receive(void *opaque)
{
  s3c2410x_serial_dev *s = opaque;
  /* If there's no byte to be read, we can receive a new one */
  return !s->rx_available;
}

static void 
s3c2410x_serial_receive(void *opaque, const uint8_t *buf, int size)
{
  s3c2410x_serial_dev *s = opaque;
  s->rx_byte = buf[0];
  s->rx_available = 1;
  if( s->ucon & 1<<8 ) {
    S3C2410X_DBF("Serial[%d] setting RX level interrupt due to incoming char %02x\n", s->port, buf[0]);
    s3c2410x_irq_set_subsrc_interrupt_level(s->port*3, 1);
  } else {
    S3C2410X_DBF("Serial[%d] pulse receive interrupt %d for %02x\n", s->port, 1<<(s->port*3), buf[0]);
    s3c2410x_irq_raise_subsrc_interrupt(s->port*3);
  }
}

static void
s3c2410x_serial_save(QEMUFile *f, void *opaque)
{
  s3c2410x_serial_dev *s = opaque;
  qemu_put_be32(f, s->ulcon);
  qemu_put_be32(f, s->ucon);
  qemu_put_be32(f, s->ufcon);
  qemu_put_be32(f, s->umcon);
  qemu_put_be32(f, s->ubrdiv);
  qemu_put_byte(f, s->rx_byte);
  qemu_put_byte(f, s->rx_available);
}

static int
s3c2410x_serial_load(QEMUFile *f, void *opaque, int version_id)
{
  s3c2410x_serial_dev *s = opaque;
  
  if (version_id != 0) return -EINVAL;
  
  s->ulcon = qemu_get_be32(f);
  s->ucon = qemu_get_be32(f);
  s->ufcon = qemu_get_be32(f);
  s->umcon = qemu_get_be32(f);
  s->ubrdiv = qemu_get_be32(f);
  s->rx_byte = qemu_get_byte(f);
  s->rx_available = qemu_get_byte(f) == 1;
  return 1;
}

void
s3c2410x_serial_init(int port)
{
  /* Initialise a serial port at the given port address */
  s3c2410x_serial_dev *s;
  int serial_io_magic;
  
  s = qemu_mallocz(sizeof(s3c2410x_serial_dev));
  if (!s) return;
  s->chr = serial_hds[port];
  s->ulcon = 0;
  s->ucon = 0;
  s->ufcon = 0;
  s->umcon = 0;
  s->ubrdiv = 0;
  s->port = port;
  s->rx_available = 0;
  /* Register ourselves for saving/loading of state */
  register_savevm("serial", port, 1, s3c2410x_serial_save, s3c2410x_serial_load, s);
  /* Prepare our MMIO tag */
  serial_io_magic = cpu_register_io_memory(0, s3c2410x_serial_read, s3c2410x_serial_write, s);
  /* Register the region with the tag */
  cpu_register_physical_memory(CPU_S3C2410X_SERIAL_BASE(port), 44, serial_io_magic);
  if (s->chr) {
    /* Add ourselves to the character device's IO handlers, if the port is there */
    qemu_chr_add_read_handler(s->chr, s3c2410x_serial_can_receive, s3c2410x_serial_receive, s);
    qemu_chr_add_event_handler(s->chr, s3c2410x_serial_event);
  }
}
