/* hw/s3c2410x_timers.c
 *
 * Samsung S3C2410X emulation
 *
 * Copyright 2006 Daniel Silverstone and Vincent Sanders
 *
 * This file is under the terms of the GNU General Public
 * License Version 2
 */

#include "vl.h"
#include "s3c2410x.h"

static uint32_t s3c2410x_timers_reg[17];
QEMUTimer *timer4;
static uint32_t timer4_reload_value;
static int64_t timer4_last_ticked;

/*
  
QEMU_TIMER_BASE is ticks per second for the qemu clocks
TCLK1 (assumed input for timer4) is 12 MHz
Thus, period in ticks of timer4 is:

(timer4_period * QEMU_TIMER_BASE) / 12000000

 */
  
#define MEGA 1000000
#define HERTZ 1
  
#define TCLK1 (12 * MEGA * HERTZ)
  
static void
s3c2410x_schedule_timer4(void)
{
  s3c2410x_timers_reg[CPU_S3C2410X_TIMERS_TCNTB4] = timer4_reload_value;
  timer4_last_ticked = qemu_get_clock(vm_clock);
  qemu_mod_timer(timer4, timer4_last_ticked + ((timer4_reload_value*ticks_per_sec)/TCLK1));
}

static void
s3c2410x_timer4_tick(void *opaque_ignored)
{
  S3C2410X_DBF("Timer4 goes TICK!\n");
  s3c2410x_irq_raise_interrupt(14);
  if (s3c2410x_timers_reg[CPU_S3C2410X_TIMERS_TCON] && (1<<22)) {
    s3c2410x_schedule_timer4();
  }
}

static void
s3c2410x_timers_write_f(void *opaque_ignored, target_phys_addr_t addr_, uint32_t value)
{
  int addr = (addr_ >> 2) & 0x1f;
  S3C2410X_DBF("Write TIMERS[%02x] = %08x\n",addr<<2, value);
  s3c2410x_timers_reg[addr] = value;
  if( addr == CPU_S3C2410X_TIMERS_TCON ) {
    /* If Timer4's manual update is set, copy in the reload value */
    if( value & (1 << 21) ) 
      timer4_reload_value = s3c2410x_timers_reg[CPU_S3C2410X_TIMERS_TCNTB4];
    /* If Timer4's manual update is unset, and the timer is running, start it */
    if( !(value & (1 << 21)) && value & (1 << 20)) {
      s3c2410x_schedule_timer4();
    }
  }
}

static uint32_t
s3c2410x_timers_read_f(void *opaque_ignored, target_phys_addr_t addr_)
{
  int addr = (addr_ >> 2) & 0x1f;
  S3C2410X_DBF("Read TIMERS[%02x]\n",addr<<2);
  if( addr == CPU_S3C2410X_TIMERS_TCNTO4 ) {
    return timer4_reload_value - (((qemu_get_clock(vm_clock)-timer4_last_ticked)*TCLK1)/ticks_per_sec);
  }
  return s3c2410x_timers_reg[addr];
}


static CPUReadMemoryFunc *s3c2410x_timers_read[] = {
  &s3c2410x_timers_read_f,
  &s3c2410x_timers_read_f,
  &s3c2410x_timers_read_f,
};

static CPUWriteMemoryFunc *s3c2410x_timers_write[] = {
  &s3c2410x_timers_write_f,
  &s3c2410x_timers_write_f,
  &s3c2410x_timers_write_f,
};

void
s3c2410x_timers_init(void)
{
  /* Samsung S3C2410X TIMERS magic badger.
   *
   * PWM timers.
   */
  int tag = cpu_register_io_memory(0, s3c2410x_timers_read, s3c2410x_timers_write, 0);
  cpu_register_physical_memory(CPU_S3C2410X_TIMERS_BASE, 17*4, tag);
  timer4 = qemu_new_timer(vm_clock, s3c2410x_timer4_tick, 0);
}

/*
  
Need to set up a timer against the vm_clock for 12MHz. (Virtual TCLK1)

Then we only provide timer4 anyway, and it will always be configured to use TCLK1

We need to take on board the count/compare values, and when the timer is reloaded, we set
a qemu_timer_mod to call us back at the appropriate VM tick.

At that point we probably raise an interrupt via s3c2410x_irq_raise_interrupt(14)

*/
